/*
Copyright (C) 2000,2001 Stefan Duffner 

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <qstring.h>
#include <qvaluelist.h>

#include "TransitionInfo.h"
#include "IOInfoBin.h"
#include "IOInfoList.h"
#include "IOInfo.h"
//#include "IOInfoASCII.h"


/// Constructor
IOInfo::IOInfo(IOType t)
{
  type = t;
  role = FALSE;
}

/// Destructor
IOInfo::~IOInfo()
{
}


/// Returns TRUE if @a io equals this object
bool IOInfo::equals(IOInfo* io)
{
  if (io->type!=type)
    return FALSE;

  IOInfoBin b1(type), b2(type);
  int len1, len2, maxlen;
  len1 = getLength();
  len2 = io->getLength();

  maxlen = len1>len2? len1 : len2;

  b1 = convertToBin(maxlen);
  b2 = io->convertToBin(maxlen);

  QString s1, s2;
  s1 = b1.convertToBinStr();
  s2 = b2.convertToBinStr();

  if (s1==s2)
    return TRUE;
  else
    return FALSE;
}


/// Returns TRUE if @a io matches this object (i.e. this I/O info contains @a io)
bool IOInfo::matches(IOInfo* io)
{
  IOInfoList list1;
  IOInfoList list2;
  list1.setAutoDelete(TRUE);
  list2.setAutoDelete(TRUE);

  convertToBinList(list1);
  io->convertToBinList(list2);

  QListIterator<IOInfo> it(list2);
  for(;it.current(); ++it)
  {
    //qDebug("%s", it.current()->convertToString().latin1());
    if (!list1.contains(it.current()))
    {
      //qDebug("not contained: %s", it.current()->convertToString().latin1());
      return FALSE;
    }
  }
  return TRUE;
}


void IOInfo::expandList(IOInfoList& input, IOInfoList& list)
{
  /*
  class IOInfoEventList : public IOInfoList
  {
    private:
      bool start;
  };
  IOInfoEventList eventlist;
  */

  // creating eventlist
  IOInfoList eventlist;
  eventlist.setAutoDelete(TRUE);
  IOInfo *io;
  IOInfo *start, *end;
  IOInfo *new_start, *new_end;
  QPtrListIterator<IOInfo> inputit(input);
  for(; inputit.current(); ++inputit)
  {
    IOInfoList iolist;
    iolist.setAutoDelete(TRUE);
    inputit.current()->split(iolist);

    QPtrListIterator<IOInfo> it(iolist);
    for(; it.current(); ++it)
    {
      io = it.current();
      //qDebug("io: %s", io->convertToString().latin1());

      QPtrList<IOInfo> list;
      list.setAutoDelete(TRUE);
      list = io->getSingles();

      if (!list.isEmpty())
      {
	start = list.getFirst();
	end = list.getLast();

	new_start = start->clone();
	new_end = end->clone();
	new_start->setRole(TRUE);
	new_end->setRole(FALSE);

	/*
	QString s,e;
	QString sc,ec;
	s = new_start->convertToString();
	e = new_end->convertToString();
	sc = new_start->convertToComparableString();
	ec = new_end->convertToComparableString();
	*/
	//qDebug("cel: start: %s %d  end: %s %d", s.latin1(), new_start->getRole(), e.latin1(), new_end->getRole());
	/*
	qDebug("cel: start: %s %d  end: %s %d",
	    new_start->convertToComparableString().latin1(),
	    new_start->getRole(), new_end->convertToComparableString().latin1(), new_end->getRole());
	    */
	/*
	if (sc>ec)
	  qDebug("s>e");
        else if (sc<ec)
	  qDebug("s<e");
	else
	  qDebug("s==e");
	eventlist.compareItems(new_start, new_end);
	*/
	
	eventlist.append(new_start);
	eventlist.append(new_end);
      }
    }
  }

  // creating output while traversing eventlist (sort of line sweep algorithm)

  int open_counter=0;  // open ranges
  //int prev_open;
  IOInfo* previo=NULL;
  IOInfo *io_minus1=NULL, *io_plus1=NULL;
  QString s, e;
  
  eventlist.sort();
  //qHeapSort(eventlist.first(), eventlist.last());
  //qHeapSort(eventlist);
  start=end=NULL;
  io = NULL;
  QPtrListIterator<IOInfo> evit(eventlist);
  for(; evit.current(); ++evit)
  {
    previo = io;
    io = evit.current();

    //qDebug("event: %s    role: %d", io->convertToString().latin1(), io->getRole());
    if (io->getRole()==TRUE)	// start event
    {
      if (open_counter==0)
	start = io;
      else
      {
	io_minus1 = io->getMinus1();
	s = start->convertToString();
	e = io_minus1->convertToString();
	if (eventlist.compareItems(start, io_minus1)<=0)
	{
	  //qDebug("start: %s  end: %s", s.latin1(), e.latin1());
	  IOInfo* new_io = start->clone();
	  new_io->setRange(start, io_minus1);
	  list.append(new_io);
	}

	delete io_minus1;
        
	start = io;
      }
      open_counter++;
    }
    else			// end event
    {
      open_counter--;
      //if (open_counter==0)
      if (!io->equals(previo) || (io->equals(previo) 
	    && previo!=NULL && io->getRole()!=previo->getRole()))
	end = io;

      if (start && end)
      {
	s = start->convertToString();
	e = end->convertToString();
	if (eventlist.compareItems(start, end)<=0)
	{
	  //qDebug("start: %s  end: %s", s.latin1(), e.latin1());
	  IOInfo* new_io = start->clone();
	  new_io->setRange(start, end);
	  list.append(new_io);
	}
      }
      if (open_counter>0)
      {
	if (io_plus1)
	  delete io_plus1;
	io_plus1 = io->getPlus1();
	start = io_plus1;
      }
    }
  }
}

bool IOInfo::expandListOld(IOInfoList& input, IOInfoList& list)
{
  bool split=FALSE;

  do
  {
    list.clear();
    split = expandListRec(input, list);

    input.clear();
    QPtrListIterator<IOInfo> it(list);
    for(;it.current(); ++it)
    {
      input.append(it.current()->clone());
    }

  qDebug("  list:");
  printf("  ");
  QListIterator<IOInfo> tmpit2(list);
  for(;tmpit2.current();++tmpit2)
  {
    IOInfo* io = tmpit2.current();
    printf("%s ", io->convertToString().latin1());
  }
  printf("\n");
  } while (split);
  return TRUE;
}

bool IOInfo::expandListRec(IOInfoList& input, IOInfoList& list)
{
  IOInfo *inputio, *io, *inputio2, *io2;
  int split_occured=0;
  bool split_occured_total=FALSE;;
  IOInfoList iolist, iolist2, res, newlist, splitted_list;
  iolist.setAutoDelete(TRUE);
  iolist2.setAutoDelete(TRUE);
  res.setAutoDelete(TRUE);
  splitted_list.setAutoDelete(TRUE);
  newlist.setAutoDelete(FALSE);
  //bool first_removed=FALSE;;

  QListIterator<IOInfo> inputit(input);

  /*
  for(; inputit.current(); ++inputit)
  {
    inputio = inputit.current();
    inputio->split(iolist);
    QPtrListIterator<IOInfo> it(iolist);
    for(;it.current();++it)
    {
      io = it.current();
      list.append(io->clone());
    }
  }
  inputit.toFirst();
  */

  /*
  qDebug("\n\n\nexpandList"); // %s", convertToString().latin1());
  qDebug("  inputlist:");
  printf("  ");
  QListIterator<IOInfo> tmpit(input);
  for(;tmpit.current();++tmpit)
  {
    io = tmpit.current();
    printf("%s ", io->convertToString().latin1());
  }
  printf("\n");
  qDebug("  list:");
  printf("  ");
  QListIterator<IOInfo> tmpit2(list);
  for(;tmpit2.current();++tmpit2)
  {
    io = tmpit2.current();
    printf("%s ", io->convertToString().latin1());
  }
  printf("\n");
  */
  
  splitted_list.clear();
  

  for(;inputit.current(); ++inputit)
  {
    inputio = inputit.current();

    //qDebug("\n\ntreating %s", inputio->convertToString().latin1());

    inputio->split(iolist);
    QPtrListIterator<IOInfo> it(iolist);
    for(;it.current();++it)
    {
      io = it.current();


      QListIterator<IOInfo> inputit2(input);
      for(;inputit2.current(); ++inputit2)
      {
	inputio2 = inputit2.current();

	inputio2->split(iolist2);
	QPtrListIterator<IOInfo> it2(iolist2);
	for(;it2.current();++it2)
	{
	  io2 = it2.current();

	  split_occured = io->split(io2, res);
	  if (split_occured)
	  {
	    //qDebug("splitted %s and %s", io->convertToString().latin1(), io2->convertToString().latin1());
	    split_occured_total=TRUE;

	    //bool found1=FALSE;
	    //bool found2=FALSE;
	    /*
	    if (split_occured==1)
	      found2=TRUE;
	    else if (split_occured==2)
	      found1=TRUE;
	      */

	    /*
            QPtrListIterator<IOInfo> ittmp(splitted_list);
	    do
	    {
	      if (!found1) //if ((split_occured==1 || split_occured==3) && !found1)
	      {
		if (ittmp.current()->equals(io))
		{
		  qDebug("  removed %s", ittmp.current()->convertToString().latin1());
		  splitted_list.remove(ittmp.current());
		  found1=TRUE;
		}
	      }
	      if (!found2) //else if ((split_occured==2 || split_occured==3) && !found2)
	      {
		if (ittmp.current()->equals(io2))
		{
		  qDebug("  removed %s", ittmp.current()->convertToString().latin1());
		  splitted_list.remove(ittmp.current());
		  found2=TRUE;
		}
	      }
	      ++ittmp;
	    } while (ittmp.current() && (!found1 || !found2));
	  */

	  }

	  QPtrListIterator<IOInfo> resit(res);

	  for(; resit.current(); ++resit)
	  {
	    if (!splitted_list.contains(resit.current()))
	    {
	      splitted_list.append(resit.current()->clone());
	    }
	  }
	}
      }
    }
  }


  QPtrListIterator<IOInfo> resit(splitted_list);
  // assign splitted_list to list
  for(; resit.current(); ++resit)
  {
    if (!list.contains(resit.current()))
      list.append(resit.current()->clone());
  }

  return split_occured_total;
}


bool IOInfo::expandListOld2(IOInfoList& input, IOInfoList& list)
{
  IOInfo *inputio, *io, *listio;
  bool split_occured=FALSE;;
  bool split_occured_total=FALSE;;
  IOInfoList iolist, res, newlist, splitted_list;
  iolist.setAutoDelete(TRUE);
  res.setAutoDelete(TRUE);
  splitted_list.setAutoDelete(TRUE);
  newlist.setAutoDelete(FALSE);
  bool first_removed=FALSE;;

  QListIterator<IOInfo> inputit(input);

  /*
  for(; inputit.current(); ++inputit)
  {
    inputio = inputit.current();
    inputio->split(iolist);
    QPtrListIterator<IOInfo> it(iolist);
    for(;it.current();++it)
    {
      io = it.current();
      list.append(io->clone());
    }
  }
  inputit.toFirst();
  */

  /*
  qDebug("\n\n\nexpandList"); // %s", convertToString().latin1());
  qDebug("  inputlist:");
  printf("  ");
  QListIterator<IOInfo> tmpit(input);
  for(;tmpit.current();++tmpit)
  {
    io = tmpit.current();
    printf("%s ", io->convertToString().latin1());
  }
  printf("\n");
  qDebug("  list:");
  printf("  ");
  QListIterator<IOInfo> tmpit2(list);
  for(;tmpit2.current();++tmpit2)
  {
    io = tmpit2.current();
    printf("%s ", io->convertToString().latin1());
  }
  printf("\n");
  */
  
  
  for(;inputit.current(); ++inputit)
  {
    inputio = inputit.current();

    //qDebug("\n\ntreating %s", inputio->convertToString().latin1());

    inputio->split(iolist);
    QListIterator<IOInfo> it(iolist);
    for(;it.current();++it)
    {
      io = it.current();

      //newlist.clear();
      splitted_list.clear();

      if (!list.isEmpty())
      {
	QListIterator<IOInfo> listit(list);
	for(;listit.current(); ++listit)
	{
	  //if (split_occured && listit.current()!=list.first())
	    //--listit;
	  if (split_occured && first_removed)
	    listit.toFirst();

	  listio = listit.current();

	  //qDebug("\nsplitting %s and %s", io->convertToString().latin1(), listio->convertToString().latin1());
	  split_occured = io->split(listio, res);

	  QListIterator<IOInfo> resit(res);

	  if (split_occured)
	  {
	    split_occured_total=TRUE;
	    //expandList(res, newlist);
	    QPtrListIterator<IOInfo> listit2(list);
	    listit2=listit;
            --listit2;

	    //qDebug("  removing %s", listio->convertToString().latin1());
	    list.remove(listio);

	    if (listit2==0)
	      first_removed=TRUE;
	    else
	    {
	      listit = listit2;
	      first_removed=FALSE;
	    }

	    //else
	    //if (listit!=0)
	      //++listit;

	    /*
	    listit2=listit;
	    ++listit2;
	    if (first_removed)
	      listit2.toFirst();
	    if (listit2.current())
	      qDebug("  next-item %s", listit2.current()->convertToString().latin1());
	      */
	    //qDebug("  removing %s", io->convertToString().latin1());
	    //list.remove(io);
	    //expandList(res, list);

	    for(; resit.current(); ++resit)
	    {
	      //qDebug("res: %s", resit.current()->convertToString().latin1());
	      //if (!newlist.contains(resit.current()))
	      if (!splitted_list.contains(resit.current()))
	      {
		//newlist.append(resit.current()->clone());
		splitted_list.append(resit.current()->clone());
		//qDebug("appending to split list: %s", resit.current()->convertToString().latin1());
	      }
	    }
	  }
	  else
	  {
	    for(; resit.current(); ++resit)
	    {
	      //qDebug("res: %s", resit.current()->convertToString().latin1());
	      //if (!newlist.contains(resit.current()))
	      if (!list.contains(resit.current()))
	      {
		//newlist.append(resit.current()->clone());
		list.append(resit.current()->clone());
		//qDebug("appending to list: %s", resit.current()->convertToString().latin1());
	      }
	    }
	  }

	}
      }  
      else   // list is empty
      {
	//newlist.append(io->clone());
	list.append(io->clone());
	//qDebug("first element %s appended", io->convertToString().latin1());
      }

      //list = newlist;


      if (!splitted_list.isEmpty())
      {
	/*
	IOInfoList splitted_list_exp;
	splitted_list_exp.setAutoDelete(TRUE);
	expandList(splitted_list, splitted_list_exp);
	*/

	//QPtrListIterator<IOInfo> resit(splitted_list_exp);
	QPtrListIterator<IOInfo> resit(splitted_list);
	// append the splitted results to the list
	for(; resit.current(); ++resit)
	{
	  if (!list.contains(resit.current()))
	    list.append(resit.current()->clone());
	}
	IOInfoList list_exp;
	list_exp.setAutoDelete(FALSE);
	// need another expand cycle for list+splitted_list_exp
	expandList(list, list_exp);
	list = list_exp;
      }


      /*
      qDebug("list:");
      QListIterator<IOInfo> listit(list);
      for(;listit.current();++listit)
      {
	io = listit.current();
	printf("%s ", io->convertToString().latin1());
      }
      printf("\n");
      */
    }
    //qDebug("end treatment");
  }
  /*
  qDebug("end expandList"); 
  qDebug("  inputlist:");
  printf("  ");
  tmpit.toFirst();
  for(;tmpit.current();++tmpit)
  {
    io = tmpit.current();
    printf("%s ", io->convertToString().latin1());
  }
  printf("\n");
  qDebug("  list:");
  printf("  ");
  tmpit2.toFirst();
  for(;tmpit2.current();++tmpit2)
  {
    io = tmpit2.current();
    printf("%s ", io->convertToString().latin1());
  }
  printf("\n");
  */
  
  return split_occured_total;
}

